/*
 * Algebraic manipulator simplifying routines.
 *
 * Copyright (c) 1996 George Gesslein II.
 */

#include "am.h"
#include "externs.h"

#define	MAX_COMPARE_TERMS	100	/* Max no. of terms on same level that can be compared. */

simp_sub(n)
int	n;
{
	simp_side(&lhs[n][0], &n_lhs[n]);
	simp_side(&rhs[n][0], &n_rhs[n]);
}

simpa_sub(n, poly_flag, quick_flag)
int	n;
int	poly_flag, quick_flag;
{
	simpa_side(&lhs[n][0], &n_lhs[n], poly_flag, quick_flag);
	simpa_side(&rhs[n][0], &n_rhs[n], poly_flag, quick_flag);
}

simp_side(equation, np)
token_type	*equation;
int		*np;
{
	simp_ssub(equation, np, 0L, 0.0, true, false, true);
}

/*
 * Simplify routine used by the "derivative" command.
 */
simpd_side(equation, np)
token_type	*equation;
int		*np;
{
	simp_ssub(equation, np, 0L, 0.0, true, true, true);
}

/*
 * This function is used by the solver.
 */
simps_side(equation, np, zsolve)
token_type	*equation;
int		*np;
int		zsolve;
{
	simp_ssub(equation, np, 0L, 0.0, !zsolve, true, true);
}

/*
 * This function is used by the "factor" command.
 */
simpv_side(equation, np, v)
token_type	*equation;
int		*np;
long		v;
{
	if (v == 0)
		simp_side(equation, np);
	else
		simp_ssub(equation, np, v, 0.0, false, false, true);
}

/*
 * Try to eliminate "i#" from an equation side,
 * by converting "i#*(b^.5)" to "(-1*b)^.5".
 */
simp_i(equation, np)
token_type	*equation;
int		*np;
{
	int	i;
	int	level;

	simp_loop(equation, np);
	uf_repeat(equation, np);
	for (i = 0; i < *np; i += 2) {
		if (equation[i].kind == VARIABLE && equation[i].token.variable == IMAGINARY) {
			if (*np + 2 > N_TOKENS) {
				error_huge();
			}
			level = equation[i].level + 1;
			blt(&equation[i+2], &equation[i], (*np - i) * sizeof(token_type));
			*np += 2;
			equation[i].level = level;
			equation[i].kind = CONSTANT;
			equation[i].token.constant = -1.0;
			i++;
			equation[i].level = level;
			equation[i].kind = OPERATOR;
			equation[i].token.operatr = POWER;
			i++;
			equation[i].level = level;
			equation[i].kind = CONSTANT;
			equation[i].token.constant = 0.5;
		}
	}
	do {
		do {
			do {
				do {
					do {
						organize(equation, np);
					} while (combine_constants(equation, np, false) > 1);
				} while (elim_k(equation, np));
			} while (simp_pp(equation, np));
		} while (fctor_power(equation, np));
	} while (factor_times(equation, np));
	simp_loop(equation, np);
}

/*
 * Main simplify code used by the "simplify" command.
 * "tlhs" and "trhs" are wiped out by the polynomial routines.
 */
simpa_side(equation, np, poly_flag, quick_flag)
token_type	*equation;
int		*np;
int		poly_flag, quick_flag;
{
	int	i;
	int	flag, flag2;

	partial_flag = true;
	simp_ssub(equation, np, 0L, 0.0, false, false, true);
	uf_power(equation, np);
	simp_ssub(equation, np, 0L, 0.0, true, false, true);
	side_debug(3, equation, *np);
	unsimp_power(equation, np);
	uf_times(equation, np);
	simp_i(equation, np);
	unsimp_power(equation, np);
	uf_times(equation, np);
	simp_ssub(equation, np, 0L, 1.0, true, false, true);
	side_debug(3, equation, *np);
	unsimp_power(equation, np);
	uf_neg_help(equation, np);
	uf_times(equation, np);
	do {
		simp_ssub(equation, np, 0L, 0.0, false, true, true);
	} while (uf_power(equation, np));
	side_debug(3, equation, *np);
	unsimp_power(equation, np);
	if (quick_flag) {
		uf_times(equation, np);
		simp_loop(equation, np);
		uf_pplus(equation, np);
	} else {
		uf_repeat(equation, np);
		uf_times(equation, np);
		factorv(equation, np, (long) IMAGINARY);
		uf_pplus(equation, np);
		uf_repeat(equation, np);
	}
	simp_ssub(equation, np, (long) MATCH_ANY, 0.0, true, false, false);
	uf_times(equation, np);
	do {
		elim_loop(equation, np);
	} while (fctor_power(equation, np));
	uf_times(equation, np);

	simp_loop(equation, np);
	uf_power(equation, np);
	do {
		simp_ssub(equation, np, 0L, 1.0, false, false, true);
	} while (uf_power(equation, np));
	unsimp2_power(equation, np);
	do {
		simp_ssub(equation, np, 0L, 1.0, false, true /* was 2 */, true);
	} while (uf_power(equation, np));
	make_fractions(equation, np);
	uf_power(equation, np);
	simp_loop(equation, np);
	unsimp_power(equation, np);
	uf_times(equation, np);
	simpb_side(equation, np);
	side_debug(2, equation, *np);
	partial_flag = false;
	flag = false;
	flag2 = false;
try_again:
	if (polydiv_simp(equation, np)) {
		flag = false;
		flag2 = false;
		simpb_side(equation, np);
		goto try_again;
	}
	if (!flag) {
		while (poly_factor(equation, np)) {
			flag = true;
			flag2 = false;
			simpb_side(equation, np);
		}
		if (flag)
			goto try_again;
	}
	if (!flag2) {
		while (div_remainder(equation, np, poly_flag)) {
			flag = false;
			flag2 = true;
			simpb_side(equation, np);
		}
		if (flag2)
			goto try_again;
	}
	partial_flag = true;
	simpb_side(equation, np);
}

typedef struct {
	long	v;
	int	count;
} sort_type;

static int
vcmp(p1, p2)
sort_type	*p1, *p2;
{
	if (((p1->v & VAR_MASK) == SIGN) == ((p2->v & VAR_MASK) == SIGN)) {
		if (p2->count == p1->count) {
			if (p1->v < p2->v)
				return -1;
			if (p1->v == p2->v)
				return 0;
			return 1;
		}
		return(p2->count - p1->count);
	} else {
		if ((p1->v & VAR_MASK) == SIGN) {
			return -1;
		} else {
			return 1;
		}
	}
}

/*
 * Neat simplify and factor routine.
 * Factors variables in order.
 * Sign variables first, then by frequency.
 */
simpb_side(equation, np)
token_type	*equation;
int		*np;
{
	int	i;
	int	vc, cnt;
	long	v1, last_v;
	sort_type	va[MAX_VARS];

	simp_loop(equation, np);
	uf_power(equation, np);
	uf_pplus(equation, np);
	vc = 0;
	last_v = -1;
	for (;;) {
		v1 = -1;
		for (i = 0; i < *np; i += 2) {
			if (equation[i].kind == VARIABLE
			    && equation[i].token.variable > last_v) {
				if (v1 == -1 || equation[i].token.variable < v1) {
					v1 = equation[i].token.variable;
					cnt = 1;
				} else if (equation[i].token.variable == v1) {
					cnt++;
				}
			}
		}
		if (v1 == -1)
			break;
		last_v = v1;
		if (vc >= ARR_CNT(va)) {
			printf("Warning: Too many variables!\n");
			break;
		}
		va[vc].v = v1;
		va[vc++].count = cnt;
	}
	if (vc == 0) {
		simp_ssub(equation, np, 0L, 0.0, true, false, true);
		return;
	}
	qsort((char *) &va[0], vc, sizeof(*va), vcmp);
	simp2_divide(equation, np, va[0].v);
	for (i = 1; i < vc; i++) {
		if (factor_divide(equation, np, va[i].v, 0.0))
			simp2_divide(equation, np, va[i].v);
	}
	simp2_divide(equation, np, 0L);
	for (i = 0; i < vc; i++) {
		while (factor_plus(equation, np, va[i].v, 0.0)) {
			simp2_divide(equation, np, 0L);
		}
	}
	while (factor_plus(equation, np, (long) MATCH_ANY, 0.0)) {
		simp2_divide(equation, np, 0L);
	}
	simp_ssub(equation, np, (long) MATCH_ANY, 0.0, true, false, true);
}

/*
 * The most basic simplification loop.
 * Just does constant simplification.
 */
elim_loop(equation, np)
token_type	*equation;
int		*np;
{
	int	count;

	count = 0;
	side_debug(6, equation, *np);
	do {
		count++;
		do {
			do {
				do {
					organize(equation, np);
				} while (combine_constants(equation, np, true) > 1);
			} while (elim_k(equation, np));
		} while (simp_pp(equation, np));
		if (count > 1)
			break;
	} while (reorder(equation, np));
	side_debug(5, equation, *np);
}

/*
 * Configurable simplify routine.
 */
simp_ssub(equation, np, v, d, power_flag, super_code, times_flag)
token_type	*equation;
int		*np;
long		v;
double		d;
int		power_flag;
int		super_code;
int		times_flag;
{
	do {
		do {
			do {
				do {
					do {
						do {
							do {
								do {
									do {
										elim_loop(equation, np);
									} while (simp2_power(equation, np));
								} while (times_flag && factor_times(equation, np));
							} while (elim_sign(equation, np));
						} while (subtract_itself(equation, np));
					} while (factor_constants(equation, np, (times_flag == 2) || (super_code == 2)));
				} while (factor_divide(equation, np, v, d));
			} while (factor_plus(equation, np, v, d));
		} while (super_code && super_factor(equation, np, super_code == 2));
	} while (power_flag && fctor_power(equation, np));
}

/*
 * Commonly used simplify routine that doesn't factor.
 */
int
simp_loop(equation, np)
token_type	*equation;
int		*np;
{
	int	i;
	int	rv;

	rv = false;
	do {
		do {
			do {
				do {
					elim_loop(equation, np);
				} while (simp2_power(equation, np));
				i = factor_times(equation, np);
				if (i)
					rv = true;
			} while (i);
		} while (elim_sign(equation, np));
	} while (subtract_itself(equation, np));
	return rv;
}

/*
 * Factor only the specified variable "v".
 */
factorv(equation, np, v)
token_type	*equation;
int		*np;
long		v;
{
	do {
		do {
			simp_loop(equation, np);
		} while (factor_plus(equation, np, v, 0.0));
	} while (v == IMAGINARY && div_imaginary(equation, np));
}

/*
 * Combine all like denominators.
 */
simp_divide(equation, np, v)
token_type	*equation;
int		*np;
long		v;
{
	do {
		do {
			simp_loop(equation, np);
		} while (factor_constants(equation, np, false));
	} while (factor_divide(equation, np, v, 0.0));
}

/*
 * Combine all like denominators.  Don't factor times.
 * No need to do complex number division.
 */
simp2_divide(equation, np, v)
token_type	*equation;
int		*np;
long		v;
{
	do {
		do {
			do {
				do {
					do {
						elim_loop(equation, np);
					} while (simp2_power(equation, np));
				} while (elim_sign(equation, np));
			} while (subtract_itself(equation, np));
		} while (factor_constants(equation, np, false));
	} while (factor_divide(equation, np, v, 0.0));
}

/*
 * Fix up levels of parentheses in an equation side.
 * Put similar operators on the same level.
 */
organize(equation, np)
token_type	*equation;
int		*np;
{
	int	eloc;

	if (*np <= 0 || (*np & 1) != 1) {
		printf("Organize called with bad expression size!  n = %d\n", *np);
		longjmp(jmp_save, 13);
	}
	if (*np > N_TOKENS) {
		printf("Memory leak detected!\n");
		side_debug(1, equation, *np);
		longjmp(jmp_save, 13);
	}
	org_recurse(equation, np, 0, 1, &eloc, 0);
}

/*
 * Convert (x^n)^m to x^(n*m).
 */
int
simp_pp(equation, np)
token_type	*equation;
int		*np;
{
	int		i, j, k;
	int		level;
	int		left_level;
	int		diff;
	int		modified;
	double		numerator, denominator;
	double		numerator2, denominator2;

	modified = false;
	for (i = 1; i < *np; i += 2) {
		if (equation[i].token.operatr == POWER) {
			level = equation[i].level;
			left_level = 1;
			for (j = i - 2; j >= 0; j -= 2) {
				if (equation[j].level < level) {
					left_level = equation[j].level;
					break;
				}
			}
			for (j = i + 2; j < *np; j += 2) {
				if (equation[j].level <= level) {
					if (left_level <= equation[j].level
					    && equation[j].token.operatr == POWER) {
						if (!symb_flag && (i + 2) == j && equation[j].level == equation[j+1].level
						    && equation[j-1].kind == CONSTANT
						    && equation[j+1].kind == CONSTANT) {
							f_to_fraction(equation[j-1].token.constant, &numerator, &denominator);
							if (fmod(numerator, 2.0) == 0.0) {
								f_to_fraction(equation[j+1].token.constant, &numerator2, &denominator2);
								if (fmod(denominator2, 2.0) == 0.0) {
									break;
								}
							}
						}
						equation[j].token.operatr = TIMES;
						for (k = i + 1; k < j; k++)
							equation[k].level++;
						diff = (level - equation[j].level) + 1;
						level = equation[j].level;
						for (k = j; k < *np; k++) {
							if (equation[k].level < level)
								break;
							equation[k].level += diff;
						}
						modified = true;
					}
					break;
				}
			}
		}
	}
	return modified;
}

/*
 * Convert x^-y to 1/(x^y).
 * Simplify k1^(k2*x).
 * Return true if equation side was modified.
 */
int
simp2_power(equation, np)
token_type	*equation;
int		*np;
{
	int	modified;
	int	i;

	modified = false;
	for (i = 1; i < *np; i += 2) {
		if (equation[i].token.operatr == POWER) {
			modified |= sp_sub(equation, np, i);
		}
	}
	if (modified)
		return true;
	for (i = 1; i < *np; i += 2) {
		if (equation[i].token.operatr == POWER) {
			modified |= sp2_sub(equation, np, i);
		}
	}
	return modified;
}

int
sp_sub(equation, np, i)
token_type	*equation;
int		*np, i;
{
	int	j, k;
	int	level;
	int	op;

	level = equation[i].level;
	op = 0;
	k = -1;
	for (j = i + 1;; j++) {
		if (j >= *np || equation[j].level < level) {
			break;
		}
		if (equation[j].level == level + 1) {
			if (equation[j].kind == OPERATOR)
				op = equation[j].token.operatr;
			else if (equation[j].kind == CONSTANT
			    && equation[j].token.constant < 0.0)
				k = j;
		}
	}
	if (j - i <= 2 && equation[i+1].kind == CONSTANT
	    && equation[i+1].token.constant < 0.0)
		k = i + 1;
	else if (k < 0)
		return false;
	switch (op) {
	case 0:
	case TIMES:
	case DIVIDE:
		if (*np + 2 > N_TOKENS) {
			error_huge();
		}
		equation[k].token.constant = -equation[k].token.constant;
		for (k = i - 2;; k--) {
			if (k < 0 || equation[k].level < level)
				break;
		}
		k++;
		for (i = k; i < j; i++)
			equation[i].level++;
		blt(&equation[k+2], &equation[k], (*np - k) * sizeof(*equation));
		*np += 2;
		equation[k].level = level;
		equation[k].kind = CONSTANT;
		equation[k].token.constant = 1.0;
		k++;
		equation[k].level = level;
		equation[k].kind = OPERATOR;
		equation[k].token.operatr = DIVIDE;
		return true;
	}
	return false;
}

int
sp2_sub(equation, np, i)
token_type	*equation;
int		*np, i;
{
	int	j;
	int	level;

	level = equation[i].level;
	if (equation[i-1].level != level || equation[i-1].kind != CONSTANT)
		return false;
	if (equation[i+1].level != level + 1 || equation[i+1].kind != CONSTANT
	    || equation[i+1].token.constant == 1.0)
		return false;
	j = i + 2;
	if (j >= *np || equation[j].level != level + 1)
		return false;
	switch (equation[j].token.operatr) {
	case TIMES:
		break;
	case DIVIDE:
		if (*np + 2 > N_TOKENS) {
			error_huge();
		}
		blt(&equation[j+2], &equation[j], (*np - j) * sizeof(*equation));
		*np += 2;
		equation[j+1].level = level + 1;
		equation[j+1].kind = CONSTANT;
		equation[j+1].token.constant = 1.0;
		break;
	default:
		return false;
	}
	equation[j].level = level;
	equation[j].token.operatr = POWER;
	equation[i-1].level++;
	equation[i].level++;
	return true;
}

int
org_recurse(equation, np, loc, level, elocp, parent_op)
token_type	*equation;
int		*np, loc, level, *elocp;
int		parent_op;
{
	token_type	*p1, *bp, *ep;
	int	first_op, sub_op;
	int	op_count, op_loc;
	int	neg_one_times;
	int	i;
	int	junk;
	int	eloc;
	int	sub_eloc;
	int	min1;
	int	invert;

	bp = &equation[loc];
	ep = &equation[*np];
	min1 = bp->level;
	for (p1 = bp + 1; p1 < ep; p1 += 2) {
		if (p1->level < min1) {
			if (p1->level < level)
				break;
			min1 = p1->level;
		}
	}
	ep = p1;
	eloc = (ep - equation) - 1;
	*elocp = eloc;
	if (eloc == loc) {
		bp->level = max(level - 1, 1);
		return 0;
	}
	if (min1 > level) {
		for (p1 = bp; p1 < ep; p1++)
			p1->level -= min1 - level;
	}
	op_count = 0;
	first_op = 0;
	for (p1 = bp + 1; p1 < ep; p1 += 2) {
		if (p1->level == level) {
			op_count++;
			if (first_op == 0) {
				first_op = p1->token.operatr;
				op_loc = p1 - equation;
			} else
				break;
		}
	}
	neg_one_times = (op_count == 1
	    && (parent_op == PLUS || parent_op == MINUS)
	    && ((op_loc - loc == 1
	    && first_op == TIMES
	    && bp->kind == CONSTANT
	    && bp->token.constant == -1.0)
	    || (eloc - op_loc == 1
	    && (first_op == TIMES || first_op == DIVIDE)
	    && (ep - 1)->kind == CONSTANT
	    && (ep - 1)->token.constant == -1.0)));
	if (neg_one_times && (op_loc - loc != 1)) {
		blt(bp + 2, bp, (eloc - loc - 1) * sizeof(*equation));
		bp->level = level;
		bp->kind = CONSTANT;
		bp->token.constant = -1.0;
		(bp + 1)->level = level;
		(bp + 1)->kind = OPERATOR;
		(bp + 1)->token.operatr = TIMES;
	}
	for (i = loc; i <= eloc; i += 2) {
		if (equation[i].level > level) {
			sub_op = org_recurse(equation, np, i, level + 1, &sub_eloc, first_op);
			switch (sub_op) {
			case PLUS:
			case MINUS:
				if (!neg_one_times && first_op != PLUS && first_op != MINUS)
					break;
				if (i - 1 >= loc && equation[i-1].token.operatr == MINUS)
					invert = true;
				else
					invert = neg_one_times;
				org_up_level(&equation[i], &equation[sub_eloc], level, invert);
				if (neg_one_times) {
					binary_parenthesize(equation, sub_eloc + 1, loc + 1);
					org_recurse(equation, np, loc, level + 1, &junk, PLUS);
					first_op = sub_op;
				}
				break;
			case TIMES:
			case DIVIDE:
				if (first_op != TIMES && first_op != DIVIDE)
					break;
				if (i - 1 >= loc && equation[i-1].token.operatr == DIVIDE)
					invert = true;
				else
					invert = false;
				org_up_level(&equation[i], &equation[sub_eloc], level, invert);
				break;
			}
			i = sub_eloc;
		}
	}
	return first_op;
}

org_up_level(bp, ep, level, invert)
token_type	*bp, *ep;
int		level, invert;
{
	for (; bp <= ep; bp++) {
		bp->level--;
		if (invert && bp->level == level && bp->kind == OPERATOR) {
			switch (bp->token.operatr) {
			case PLUS:
				bp->token.operatr = MINUS;
				break;
			case MINUS:
				bp->token.operatr = PLUS;
				break;
			case TIMES:
				bp->token.operatr = DIVIDE;
				break;
			case DIVIDE:
				bp->token.operatr = TIMES;
				break;
			}
		}
	}
}

/*
 * Combine two or more constants on the same level.
 * If "iflag" is false, don't produce imaginary numbers.
 * Return true if equation side was modified.
 * Return greater than 1 if any imaginary numbers were produced.
 */
int
combine_constants(equation, np, iflag)
token_type	*equation;
int		*np;
int		iflag;
{
	return const_recurse(equation, np, 0, 1, iflag);
}

int
const_recurse(equation, np, loc, level, iflag)
token_type	*equation;
int		*np, loc, level, iflag;
{
	int		loc1;
	int		old_loc;
	int		const_count;
	int		zero;
	int		modified;
	double		d;
	double		d2, numerator, denominator;
	k_type		k1;
	int		op;
	int		fflag;

	modified = false;
	zero = 0;
	old_loc = loc;
	const_count = 0;
	for (;; loc++) {
cheat:
		if (loc >= *np || equation[loc].level < level) {
			if (loc - old_loc == 1)
				equation[old_loc].level = max(level - 1, 1);
			return modified;
		}
		if (equation[loc].level > level) {
			modified |= const_recurse(equation, np, loc, level + 1, iflag);
			for (; loc < *np && equation[loc].level > level; loc++)
				;
			goto cheat;
		}
		if (equation[loc].kind == CONSTANT) {
			if (const_count == 0) {
				loc1 = loc;
				const_count++;
			} else {
				op = equation[loc-1].token.operatr;
				d = equation[loc1].token.constant;
				d2 = equation[loc].token.constant;
				fflag = f_to_fraction(d2, &numerator, &denominator);
				if (fflag) {
					d2 = numerator / denominator;
				}
				errno = 0;
				if (calc((loc1 <= old_loc) ? &zero : &equation[loc1-1].token.operatr,
				    &d, op, d2)) {
					if (op == POWER
					    && !domain_check) {
						if (!fflag)
							return modified;
						d = pow(-d, d2);
						check_err();
						if (!always_positive(denominator)) {
							if (!always_positive(numerator)) {
								d = -d;
							}
							goto not_imaginary;
						}
						if (!iflag)
							return modified;
						k1 = d;
						if (d2 != 0.5) {
							if (*np + 2 > N_TOKENS) {
								error_huge();
							}
							blt(&equation[loc1+2], &equation[loc1],
							    (*np - loc1) * sizeof(*equation));
							*np += 2;
						}
						equation[loc1].level = level;
						equation[loc1].kind = CONSTANT;
						equation[loc1].token.constant = k1;
						equation[loc1+1].level = level;
						equation[loc1+1].kind = OPERATOR;
						equation[loc1+1].token.operatr = TIMES;
						equation[loc].level = level + 1;
						equation[loc].kind = VARIABLE;
						equation[loc].token.variable = IMAGINARY;
						if (d2 != 0.5) {
							equation[loc+1].level = level + 1;
							equation[loc+1].kind = OPERATOR;
							equation[loc+1].token.operatr = POWER;
							equation[loc+2].level = level + 1;
							equation[loc+2].kind = CONSTANT;
							equation[loc+2].token.constant = d2 * 2.0;
						}
						return 2;
					} else {
not_imaginary:
						k1 = d;
						equation[loc1].token.constant = k1;
						modified |= 1;
						domain_check = false;
						blt(&equation[loc-1], &equation[loc+1],
						    (*np - (loc + 1)) * sizeof(*equation));
						*np -= 2;
						loc -= 2;
					}
				}
			}
		}
	}
}

int
calc(op1p, k1p, op2, k2)
int	*op1p;
double	*k1p;
int	op2;
double	k2;
{
	int	op1;
	double	d;
	double	d1;

	op1 = *op1p;
	switch (op2) {
	case PLUS:
	case MINUS:
		if (op1 == 0)
			op1 = PLUS;
		if (op1 == PLUS)
			d = *k1p;
		else
			d = -(*k1p);
#if	EPSILON
		d1 = fabs(d);
		d1 *= epsilon;
#endif
		if (op2 == PLUS) {
			d += k2;
		} else {
			d -= k2;
		}
#if	EPSILON
		if (fabs(d) < d1)
			d = 0.0;
#endif
		if (*op1p == 0)
			*k1p = d;
		else {
			if (d >= 0.0) {
				*op1p = PLUS;
				*k1p = d;
			} else {
				*op1p = MINUS;
				*k1p = -d;
			}
		}
		break;
	case TIMES:
	case DIVIDE:
		if (op1 == 0)
			op1 = TIMES;
		if (op1 == op2) {
			*k1p *= k2;
		} else {
			if (op1 == DIVIDE) {
				*k1p = k2 / *k1p;
				*op1p = TIMES;
			} else if (op2 == DIVIDE) {
				*k1p = *k1p / k2;
			}
		}
		break;
	case POWER:
		if (*k1p == 0.0 && k2 < 0.0) {
			printf("Divide by zero error!\n");
			longjmp(jmp_save, 2);
		}
		domain_check = true;
		if (*k1p < 0.0 && fmod(k2, 1.0) != 0.0) {
			domain_check = false;
			break;
		}
		if (*k1p == 0.0 && k2 == 0.0)
			d = 1.0;	/* some people don't know 0^0 */
		else
			d = pow(*k1p, k2);
		check_err();
		if (domain_check)
			*k1p = d;
		break;
	case FACTORIAL:
		if (*k1p < 0.0) {
			printf("Factorial not defined for negative numbers!\n");
			longjmp(jmp_save, 2);
		}
		if (fmod(*k1p, 1.0) != 0.0) {
			printf("Factorial operand not integer!\n");
			longjmp(jmp_save, 2);
		}
		if (*k1p > 170.0) {
			printf("Floating point overflow!\n");
			longjmp(jmp_save, 2);
		}
		d = 1.0;
		for (d1 = 2.0; d1 <= *k1p; d1 += 1.0) {
			d *= d1;
		}
		*k1p = d;
		break;
	default:
		return false;
	}
	return true;
}

/*
 * Eliminate operations that involve 0 or 1 that can be simplified.
 * Fix addition/subtraction of negative constants and variables.
 * Replace division by a constant with times its reciprocal.
 * Return true if one or more eliminations occurred.
 */
int
elim_k(equation, np)
token_type	*equation;
int		*np;
{
	token_type	*p1, *p2;
	token_type	*ep;
	int		modified;
	int		level;
	double		d;
	double		numerator, denominator;
	k_type		k1;

	modified = false;
	for (p1 = equation + 1;;) {
		ep = &equation[*np];
		if (p1 >= ep)
			break;
		if (p1->kind == OPERATOR) {
			level = p1->level;
			switch (p1->token.operatr) {
			case PLUS:
			case MINUS:
				if (p1 + 2 < ep && (p1 + 2)->level == level + 1
				    && ((p1 + 2)->token.operatr == TIMES || (p1 + 2)->token.operatr == DIVIDE)
				    && (p1 + 1)->kind == CONSTANT
				    && (p1 + 1)->token.constant < 0.0) {
					if (p1->token.operatr == PLUS)
						p1->token.operatr = MINUS;
					else
						p1->token.operatr = PLUS;
					(p1 + 1)->token.constant = -((p1 + 1)->token.constant);
				}
				if ((p1 + 1)->level == level && (p1 + 1)->kind == CONSTANT) {
					if ((p1 + 1)->token.constant < 0.0) {
						if (p1->token.operatr == PLUS)
							p1->token.operatr = MINUS;
						else
							p1->token.operatr = PLUS;
						(p1 + 1)->token.constant = -(p1 + 1)->token.constant;
					}
					if ((p1 + 1)->token.constant == 0.0) {
						blt(p1, p1 + 2, (int) ((char *) ep - (char *) (p1 + 2)));
						*np -= 2;
						modified = true;
						continue;
					}
				}
				break;
			}
			switch (p1->token.operatr) {
			case PLUS:
				p2 = p1 - 1;
				if (p2->level == level && p2->kind == CONSTANT
				    && p2->token.constant == 0.0) {
					blt(p2, p1 + 1, (int) ((char *) ep - (char *) (p1 + 1)));
					*np -= 2;
					modified = true;
					continue;
				}
				break;
			case MINUS:
				p2 = p1 - 1;
				if (p2->level == level && p2->kind == CONSTANT
				    && p2->token.constant == 0.0) {
					if (p2 == equation
					    || (p2 - 1)->level < level) {
						p2->token.constant = -1.0;
						p1->token.operatr = TIMES;
						binary_parenthesize(equation, *np, p1 - equation);
						modified = true;
						continue;
					}
				}
				break;
			case TIMES:
				if ((p1 - 1)->level == level && (p1 - 1)->kind == CONSTANT) {
					if ((p1 - 1)->token.constant == 0.0) {
						for (p2 = p1 + 2; p2 < ep; p2 += 2) {
							if (p2->level < level)
								break;
						}
						blt(p1, p2, (int) ((char *) ep - (char *) p2));
						*np -= p2 - p1;
						modified = true;
						continue;
					}
#if	EPSILON
					if (fabs((p1 - 1)->token.constant - 1.0) <= epsilon) {
#else
					if ((p1 - 1)->token.constant == 1.0) {
#endif
						blt(p1 - 1, p1 + 1, (int) ((char *) ep - (char *) (p1 + 1)));
						*np -= 2;
						modified = true;
						continue;
					}
				}
				if ((p1 + 1)->level == level && (p1 + 1)->kind == CONSTANT) {
					d = (p1 + 1)->token.constant;
					if (d == 0.0) {
						for (p2 = p1 - 1; p2 > &equation[0]; p2--) {
							if ((p2 - 1)->level < level)	/* No mistake.  Wipe out level! */
								break;
						}
						blt(p2, p1 + 1, (int) ((char *) ep - (char *) (p1 + 1)));
						*np -= (p1 + 1) - p2;
						p1 = p2 + 1;
						modified = true;
						continue;
					}
#if	EPSILON
					if (fabs(d - 1.0) <= epsilon) {
#else
					if (d == 1.0) {
#endif
						blt(p1, p1 + 2, (int) ((char *) ep - (char *) (p1 + 2)));
						*np -= 2;
						modified = true;
						continue;
					}
					for (p2 = p1 - 1; p2 > &equation[0]; p2--) {
						if ((p2 - 1)->level < level)
							break;
					}
					if (p2->level == level && p2->kind == CONSTANT) {
						break;
					}
					blt(p2 + 2, p2, (int) ((char *) p1 - (char *) p2));
					p2->level = level;
					p2->kind = CONSTANT;
					p2->token.constant = d;
					(p2 + 1)->level = level;
					(p2 + 1)->kind = OPERATOR;
					(p2 + 1)->token.operatr = TIMES;
					if (p2 > &equation[0]) {
						p1 = p2 - 1;
						continue;
					}
				}
				break;
			case DIVIDE:
				p2 = p1 - 1;
				if (p2->level == level && p2->kind == CONSTANT
				    && p2->token.constant == 0.0) {
					for (p2 = p1 + 2; p2 < ep; p2++) {
						if (p2->level < level)
							break;
					}
					blt(p1, p2, (int) ((char *) ep - (char *) p2));
					*np -= p2 - p1;
					modified = true;
					continue;
				}
				p2 = p1 + 1;
				if (p2->level == level && p2->kind == CONSTANT) {
					if (f_to_fraction(p2->token.constant, &numerator, &denominator)) {
						p2->token.constant = denominator / numerator;
					} else {
						k1 = 1.0 / p2->token.constant;
						p2->token.constant = k1;
					}
					p1->token.operatr = TIMES;
					continue;
				}
				if (p2->level == level
				    && p2->kind == VARIABLE && (p2->token.variable & VAR_MASK) == SIGN) {
					p1->token.operatr = TIMES;
					continue;
				}
				break;
			case POWER:
				if ((p1 - 1)->level == level && (p1 - 1)->kind == CONSTANT) {
					if ((p1 - 1)->token.constant == 1.0) {
						for (p2 = p1 + 2; p2 < ep; p2++) {
							if (p2->level <= level)
								break;
						}
						blt(p1, p2, (int) ((char *) ep - (char *) p2));
						*np -= p2 - p1;
						modified = true;
						continue;
					}
				}
				if ((p1 + 1)->level == level && (p1 + 1)->kind == CONSTANT) {
					if ((p1 + 1)->token.constant == 0.0) {
						for (p2 = p1 - 1; p2 > &equation[0]; p2--) {
							if ((p2 - 1)->level <= level)
								break;
						}
						blt(p2, p1 + 1, (int) ((char *) ep - (char *) (p1 + 1)));
						*np -= (p1 + 1) - p2;
						p2->token.constant = 1.0;
						p1 = p2 + 1;
						modified = true;
						continue;
					}
#if	EPSILON
					if (fabs((p1 + 1)->token.constant - 1.0) <= epsilon) {
#else
					if ((p1 + 1)->token.constant == 1.0) {
#endif
						blt(p1, p1 + 2, (int) ((char *) ep - (char *) (p1 + 2)));
						*np -= 2;
						modified = true;
						continue;
					}
				}
				break;
			}
			p1 += 2;
		} else
			p1++;
	}
	return modified;
}

/*
 * Compare two sub-expressions.
 * If equal, return true with *diff_signp = 0.
 * if p1 * -1 == p2, return true with *diff_signp = 1.
 * Otherwise return false.
 * Both expressions must be "organize"d and simplified as much as possible.
 */
int
se_compare(p1, n1, p2, n2, diff_signp)
token_type	*p1;
int		n1;
token_type	*p2;
int		n2, *diff_signp;
{
	int	l1, l2;

	l1 = min_level(p1, n1);
	l2 = min_level(p2, n2);
	return compare_recurse(p1, n1, l1, p2, n2, l2, diff_signp);
}

int
compare_recurse(p1, n1, l1, p2, n2, l2, diff_signp)
token_type	*p1;
int		n1, l1;
token_type	*p2;
int		n2, l2, *diff_signp;
{
	token_type	*pv1, *ep1, *ep2;
	int		i, j;
	int		len;
	int		first;
	int		oc2;		/* operand count 2 */
	token_type	*opa2[MAX_COMPARE_TERMS];	/* operand pointer array 2 */
	char		used[MAX_COMPARE_TERMS];	/* operand used flag array 2 */
	int		last_op1, op1, op2;
	int		diff_op;
	double		d1, c1, c2;

	*diff_signp = false;
	if (n1 <= 0 || n2 <= 0 || (n1 & 1) != 1 || (n2 & 1) != 1) {
		printf("Programming error in call to compare_recurse()!\n");
		longjmp(jmp_save, 13);
	}
	if (n1 == 1 && n2 == 1) {
		if (p1->kind != p2->kind)
			return false;
		if (p1->kind == VARIABLE) {
			if (sign_flag && (p1->token.variable & VAR_MASK) == SIGN) {
				return((p2->token.variable & VAR_MASK) == SIGN);
			} else {
				return(p1->token.variable == p2->token.variable);
			}
		} else if (p1->kind == CONSTANT) {
			if (p1->token.constant == p2->token.constant) {
				return true;
			} else if (p1->token.constant == -p2->token.constant) {
				*diff_signp = true;
				return true;
			}
#if	EPSILON
			c1 = p1->token.constant;
			c2 = p2->token.constant;
			d1 = fabs(c1);
			d1 *= epsilon;
			if (fabs(c1 - c2) < d1) {
				return true;
			}
			if (fabs(c1 + c2) < d1) {
				*diff_signp = true;
				return true;
			}
#endif
		}
		return false;
	}
	ep1 = &p1[n1];
	ep2 = &p2[n2];
	op1 = 0;
	for (pv1 = p1 + 1; pv1 < ep1; pv1 += 2) {
		if (pv1->level == l1) {
			op1 = pv1->token.operatr;
			break;
		}
	}
	op2 = 0;
	for (pv1 = p2 + 1; pv1 < ep2; pv1 += 2) {
		if (pv1->level == l2) {
			op2 = pv1->token.operatr;
			break;
		}
	}
	diff_op = false;
	if (op2 == 0) {
		if (op1 != TIMES && op1 != DIVIDE)
			return false;
		goto no_op2;
	}
	switch (op1) {
	case PLUS:
	case MINUS:
		if (op2 != PLUS && op2 != MINUS)
			diff_op = true;
		break;
	case 0:
		if (op2 != TIMES && op2 != DIVIDE)
			return false;
		break;
	case TIMES:
	case DIVIDE:
		if (op2 != TIMES && op2 != DIVIDE)
			diff_op = true;
		break;
	default:
		if (op2 != op1)
			diff_op = true;
		break;
	}
	if (diff_op) {
		if (p1->kind == CONSTANT
		    && p1->level == l1
		    && op1 == TIMES) {
#if	EPSILON
			if (fabs(fabs(p1->token.constant) - 1.0) <= epsilon) {
#else
			if (fabs(p1->token.constant) == 1.0) {
#endif
				if (!compare_recurse(p1 + 2, n1 - 2, min_level(p1 + 2, n1 - 2),
				    p2, n2, l2, diff_signp))
					return false;
				if (p1->token.constant < 0.0) {
					*diff_signp ^= true;
				}
				return true;
			}
		}
		if (p2->kind == CONSTANT
		    && p2->level == l2
		    && op2 == TIMES) {
#if	EPSILON
			if (fabs(fabs(p2->token.constant) - 1.0) <= epsilon) {
#else
			if (fabs(p2->token.constant) == 1.0) {
#endif
				if (!compare_recurse(p1, n1, l1,
				    p2 + 2, n2 - 2, min_level(p2 + 2, n2 - 2), diff_signp))
					return false;
				if (p2->token.constant < 0.0) {
					*diff_signp ^= true;
				}
				return true;
			}
		}
		return false;
	}
no_op2:
	opa2[0] = p2;
	used[0] = false;
	oc2 = 1;
	for (pv1 = p2 + 1; pv1 < ep2; pv1 += 2) {
		if (pv1->level == l2) {
			opa2[oc2] = pv1 + 1;
			used[oc2] = false;
			if (++oc2 >= ARR_CNT(opa2)) {
/*
				printf("Expression too big to compare!\n");
*/
				return false;
			}
		}
	}
	opa2[oc2] = pv1 + 1;
	last_op1 = 0;
	first = true;
	for (pv1 = p1;;) {
		for (len = 1; &pv1[len] < ep1; len += 2)
			if (pv1[len].level <= l1)
				break;
		for (i = 0;; i++) {
			if (i >= oc2) {
				if ((op1 == TIMES || op1 == DIVIDE)
				    && pv1->level == l1 && pv1->kind == CONSTANT) {
#if	EPSILON
					if (fabs(pv1->token.constant + 1.0) <= epsilon) {
#else
					if (pv1->token.constant == -1.0) {
#endif
						*diff_signp ^= true;
						break;
					}
#if	EPSILON
					if (fabs(pv1->token.constant - 1.0) <= epsilon) {
#else
					if (pv1->token.constant == 1.0) {
#endif
						break;
					}
				}
				return false;
			}
			if (used[i]) {
				continue;
			}
			switch (op1) {
			case PLUS:
			case MINUS:
				break;
			case 0:
			case TIMES:
			case DIVIDE:
				if ((last_op1 == DIVIDE) != ((i != 0)
				    && ((opa2[i] - 1)->token.operatr == DIVIDE)))
					continue;
				break;
			default:
				if ((last_op1 == 0) != (i == 0))
					return false;
				break;
			}
			if (compare_recurse(pv1, len, (pv1->level <= l1) ? l1 : (l1 + 1),
			    opa2[i], (opa2[i+1] - opa2[i]) - 1, (opa2[i]->level <= l2) ? l2 : (l2 + 1), &j)) {
				switch (op1) {
				case 0:
				case TIMES:
				case DIVIDE:
					*diff_signp ^= j;
					break;
				case PLUS:
				case MINUS:
					if (last_op1 == MINUS)
						j = !j;
					if (i != 0 && (opa2[i] - 1)->token.operatr == MINUS)
						j = !j;
					if (!first) {
						if (*diff_signp != j)
							continue;
					} else {
						*diff_signp = j;
						first = false;
					}
					break;
				default:
					if (j)
						continue;
					break;
				}
				used[i] = true;
				break;
			}
		}
		pv1 += len;
		if (pv1 >= ep1)
			break;
		last_op1 = pv1->token.operatr;
		pv1++;
	}
	for (i = 0; i < oc2; i++) {
		if (!used[i]) {
			if ((op2 == TIMES || op2 == DIVIDE)
			    && opa2[i]->level == l2 && opa2[i]->kind == CONSTANT) {
#if	EPSILON
				if (fabs(opa2[i]->token.constant + 1.0) <= epsilon) {
#else
				if (opa2[i]->token.constant == -1.0) {
#endif
					*diff_signp ^= true;
					continue;
				}
#if	EPSILON
				if (fabs(opa2[i]->token.constant - 1.0) <= epsilon) {
#else
				if (opa2[i]->token.constant == 1.0) {
#endif
					continue;
				}
			}
			return false;
		}
	}
	return true;
}

#if	false
/*
 * Substitute a/a with 1.
 * Return true if one or more substitutions occurred.
 */
int
div_itself(equation, np)
token_type	*equation;
int		*np;
{
	return div_recurse(equation, np, 0, 1);
}

int
div_recurse(equation, np, loc, level)
token_type	*equation;
int		*np, loc, level;
{
	int	modified;
	int	i, j, k;
	int	last_op2;
	int	len1, len2;
	int	diff_sign;

	modified = false;
	for (i = loc;;) {
		if (i >= *np || equation[i].level < level)
			break;
		if (equation[i].level > level) {
			modified |= div_recurse(equation, np, i, level + 1);
			i++;
			for (; i < *np && equation[i].level > level; i += 2)
				;
			continue;
		}
		i++;
	}
	for (i = loc + 1;; i += 2) {
		if (i >= *np || equation[i].level < level)
			break;
		if (equation[i].kind != OPERATOR) {
			printf("Optimization error.  (div_itself())\n");
			break;
		}
		if (equation[i].level == level) {
			if (equation[i].token.operatr == DIVIDE) {
				for (k = i + 2;; k++)
					if (k >= *np || equation[k].level <= level)
						break;
				len1 = k - (i + 1);
				last_op2 = 0;
				for (j = loc;; j++) {
					if (j >= *np || equation[j].level < level)
						break;
					if (equation[j].level == level && equation[j].kind == OPERATOR) {
						last_op2 = equation[j].token.operatr;
						continue;
					}
					if (equation[j].level >= level && last_op2 != DIVIDE) {
						for (k = j + 1;; k++)
							if (k >= *np || equation[k].level <= level)
								break;
						len2 = k - j;
						if (se_compare(&equation[i+1], len1, &equation[j], len2, &diff_sign)) {
							blt(&equation[i], &equation[i+1+len1], (*np - (i + 1 + len1)) * sizeof(*equation));
							*np -= len1 + 1;
							if (i + 1 > j) {
								i -= len2 - 1;
							} else {
								j -= len1 + 1;
							}
							blt(&equation[j+1], &equation[j+len2], (*np - (j + len2)) * sizeof(*equation));
							*np -= len2 - 1;
							equation[j].level = level;
							equation[j].kind = CONSTANT;
							if (diff_sign)
								equation[j].token.constant = -1.0;
							else
								equation[j].token.constant = 1.0;
							modified = true;
							i -= 2;
							break;
						}
						last_op2 = DIVIDE;
					}
				}

			} else if (equation[i].token.operatr != TIMES)
				break;
		}
	}
	return modified;
}
#endif

/*
 * Take out meaningless "sign" variables and negative constants.
 * Simplify imaginary numbers raised to the power of some constants
 * (including simple division ((i#^-1) -> (-1*i#))).
 * Return true if equation side was modified.
 */
int
elim_sign(equation, np)
token_type	*equation;
int		*np;
{
	int		i, j;
	int		level;
	int		modified;
	int		op;
	double		d;
	double		numerator, denominator;

	modified = false;
	for (i = 1; i < *np; i += 2) {
		level = equation[i].level;
		if (equation[i].token.operatr == DIVIDE) {
			if (equation[i+1].level == level
			    && equation[i+1].kind == VARIABLE
			    && equation[i+1].token.variable == IMAGINARY) {
				if (*np + 2 > N_TOKENS) {
					error_huge();
				}
				blt(&equation[i+2], &equation[i], (*np - i) * sizeof(token_type));
				*np += 2;
				equation[i].level = level;
				equation[i].kind = OPERATOR;
				equation[i].token.operatr = TIMES;
				i++;
				equation[i].level = level;
				equation[i].kind = CONSTANT;
				equation[i].token.constant = -1.0;
				i++;
				equation[i].level = level;
				equation[i].kind = OPERATOR;
				equation[i].token.operatr = TIMES;
				modified = true;
			}
		}
		if (equation[i].token.operatr == POWER
		    && equation[i+1].level == level
		    && equation[i+1].kind == CONSTANT) {
			f_to_fraction(equation[i+1].token.constant, &numerator, &denominator);
			if (always_positive(numerator)) {
				if (equation[i-1].level == level
				    && equation[i-1].kind == VARIABLE
				    && equation[i-1].token.variable == IMAGINARY) {
					equation[i-1].kind = CONSTANT;
					equation[i-1].token.constant = -1.0;
					equation[i+1].token.constant /= 2.0;
					modified = true;
					continue;
				}
				op = 0;
				for (j = i - 1; (j >= 0) && equation[j].level >= level; j--) {
					if (equation[j].level <= level + 1) {
						if (equation[j].kind == OPERATOR) {
							op = equation[j].token.operatr;
							break;
						}
					}
				}
				switch (op) {
				case 0:
				case TIMES:
				case DIVIDE:
					for (j = i - 1; (j >= 0) && equation[j].level >= level; j--) {
						if (equation[j].level <= level + 1) {
							if (equation[j].kind == VARIABLE
							    && (equation[j].token.variable & VAR_MASK) == SIGN) {
								equation[j].kind = CONSTANT;
								equation[j].token.constant = 1.0;
								modified = true;
							} else if (equation[j].kind == CONSTANT
							    && equation[j].token.constant < 0.0) {
								equation[j].token.constant = -equation[j].token.constant;
								modified = true;
							}
						}
					}
				}
			} else {
				if (equation[i-1].level == level
				    && equation[i-1].kind == VARIABLE) {
					if (equation[i-1].token.variable == IMAGINARY) {
						d = fmod(equation[i+1].token.constant, 4.0);
						if (d == 1.0) {
							equation[i].token.operatr = TIMES;
							equation[i+1].token.constant = 1.0;
							modified = true;
						} else if (d == 3.0) {
							equation[i].token.operatr = TIMES;
							equation[i+1].token.constant = -1.0;
							modified = true;
						}
					} else if ((equation[i-1].token.variable & VAR_MASK) == SIGN) {
						d = fmod(equation[i+1].token.constant, 2.0);
						if (d == 1.0) {
							equation[i+1].token.constant = 1.0;
							modified = true;
						}
					}
				}
			}
		}
	}
	return modified;
}

/*
 * Move the imaginary number from the denominator to the numerator.
 * Also does complex number division.
 * This often complicates expressions, so use with care.
 * Return true if equation side was modified.
 */
int
div_imaginary(equation, np)
token_type	*equation;
int		*np;
{
	int		i, j, k;
	int		n;
	int		level;
	int		ilevel;
	int		op;
	int		iloc, biloc, eiloc;
	int		eloc;

	for (i = 1; i < *np; i += 2) {
		if (equation[i].token.operatr == DIVIDE) {
			level = equation[i].level;
			op = 0;
			iloc = -1;
			biloc = -1;
			eiloc = -1;
			k = i;
			for (j = i + 1; j < *np && equation[j].level > level; j++) {
				if (equation[j].kind == OPERATOR
				    && equation[j].level == level + 1) {
					op = equation[j].token.operatr;
					k = j;
					if (iloc >= 0 && eiloc < 0) {
						eiloc = j;
					}
				} else if (equation[j].kind == VARIABLE
				    && equation[j].token.variable == IMAGINARY) {
					if (iloc >= 0) {
						op = 0;
						break;
					}
					iloc = j;
					biloc = k + 1;
				}
			}
			eloc = j;
			if (iloc >= 0 && eiloc < 0) {
				eiloc = j;
			}
			if (iloc < 0 || (op != PLUS && op != MINUS))
				continue;
			ilevel = equation[iloc].level;
			if (ilevel != level + 1) {
				if (ilevel != level + 2) {
					continue;
				}
				if (iloc > biloc
				    && equation[iloc-1].token.operatr != TIMES)
					continue;
				if (iloc + 1 < eiloc) {
					switch (equation[iloc+1].token.operatr) {
					case TIMES:
					case DIVIDE:
						break;
					default:
						continue;
					}
				}
			}
			if ((eloc - (i + 1)) + 5 + (eiloc - biloc) + *np + 2 > N_TOKENS)
				error_huge();
			n = eloc - (i + 1);
			blt(&scratch[0], &equation[i+1], n * sizeof(token_type));
			scratch[iloc-(i+1)].kind = CONSTANT;
			scratch[iloc-(i+1)].token.constant = 0.0;
			for (j = 0; j < n; j++)
				scratch[j].level += 2;
			scratch[n].level = level + 2;
			scratch[n].kind = OPERATOR;
			scratch[n].token.operatr = POWER;
			n++;
			scratch[n].level = level + 2;
			scratch[n].kind = CONSTANT;
			scratch[n].token.constant = 2.0;
			n++;
			scratch[n].level = level + 1;
			scratch[n].kind = OPERATOR;
			scratch[n].token.operatr = PLUS;
			n++;
			blt(&scratch[n], &equation[biloc], (eiloc - biloc) * sizeof(token_type));
			j = n;
			n += (eiloc - biloc);
			for (k = j; k < n; k++)
				scratch[k].level += 2;
			scratch[n].level = level + 2;
			scratch[n].kind = OPERATOR;
			scratch[n].token.operatr = POWER;
			n++;
			scratch[n].level = level + 2;
			scratch[n].kind = CONSTANT;
			scratch[n].token.constant = 2.0;
			n++;
			scratch[j+(iloc-biloc)].kind = CONSTANT;
			scratch[j+(iloc-biloc)].token.constant = 1.0;
			blt(&equation[iloc+2], &equation[iloc], (*np - iloc) * sizeof(token_type));
			*np += 2;
			equation[iloc].level = ilevel + 1;
			equation[iloc].kind = CONSTANT;
			equation[iloc].token.constant = -1.0;
			iloc++;
			equation[iloc].level = ilevel + 1;
			equation[iloc].kind = OPERATOR;
			equation[iloc].token.operatr = TIMES;
			iloc++;
			equation[iloc].level = ilevel + 1;
			blt(&equation[i+1+n], &equation[i], (*np - i) * sizeof(token_type));
			*np += n + 1;
			blt(&equation[i+1], &scratch[0], n * sizeof(token_type));
			equation[i+1+n].token.operatr = TIMES;
			return true;
		}
	}
	return false;
}

/*
 * Convert 1/m*(-1*b+y) to (y-b)/m.
 * Returns true if equation side was modified.
 * "elim_k" should be called after this if it returns true.
 *
 * This routine also checks the validity of the equation side.
 */
int
reorder(equation, np)
token_type	*equation;
int		*np;
{
	return(order_recurse(equation, np, 0, 1));
}

int
order_recurse(equation, np, loc, level)
token_type	*equation;
int		*np, loc, level;
{
	int	i, j, k, n;
	int	op;
	int	modified;

	modified = false;
	op = 0;
	if ((loc & 1) != 0) {
		printf("Expression is corrupt!  (Parenthesis incorrect)\n");
		longjmp(jmp_save, 13);
	}
	for (i = loc;;) {
		if (i >= *np)
			break;
		if (equation[i].level < level) {
			if (equation[i].kind != OPERATOR) {
				printf("Expression is corrupt!  (Parenthesis incorrect)\n");
				longjmp(jmp_save, 13);
			}
			break;
		}
		if (equation[i].level > level) {
			modified |= order_recurse(equation, np, i, level + 1);
			i++;
			for (; i < *np && equation[i].level > level; i++)
				;
			continue;
		} else {
			if (((i & 1) == 1) != (equation[i].kind == OPERATOR)) {
				printf("Expression is corrupt!  (operator in wrong place)\n");
				longjmp(jmp_save, 13);
			}
			if (equation[i].kind == OPERATOR) {
				if (op) {
					switch (equation[i].token.operatr) {
					case PLUS:
					case MINUS:
						if (op == PLUS || op == MINUS)
							break;
						op = 0;
					case TIMES:
					case DIVIDE:
						if (op == TIMES || op == DIVIDE)
							break;
					default:
						printf("Expression is corrupt!  (Incompatible operators on same level)\n");
						longjmp(jmp_save, 13);
					}
				} else {
					op = equation[i].token.operatr;
				}
			}
		}
		i++;
	}
	switch (op) {
	case PLUS:
	case MINUS:
		if (equation[loc].kind == CONSTANT && equation[loc].token.constant < 0.0) {
			if (equation[loc].level == level
			    || (equation[loc+1].level == level + 1
			    && (equation[loc+1].token.operatr == TIMES
			    || equation[loc+1].token.operatr == DIVIDE))) {
				for (j = loc + 1; j < i; j += 2) {
					if (equation[j].level == level
					    && equation[j].token.operatr == PLUS) {
						swap(equation, np, level, loc, j + 1);
						modified = true;
						break;
					}
				}
			}
		}
		break;
	case TIMES:
	case DIVIDE:
		for (j = loc + 1;; j += 2) {
			if (j >= i)
				return modified;
			if (equation[j].level == level && equation[j].token.operatr == DIVIDE)
				break;
		}
		for (k = j + 2; k < i;) {
			if (equation[k].level == level && equation[k].token.operatr == TIMES) {
				for (n = k + 2; n < i; n += 2)
					if (equation[n].level <= level)
						break;
				n -= k;
				blt(&scratch[0], &equation[k], n * sizeof(*equation));
				blt(&equation[j+n], &equation[j], (k - j) * sizeof(*equation));
				blt(&equation[j], &scratch[0], n * sizeof(*equation));
				j += n;
				k += n;
				modified = true;
				continue;
			}
			k += 2;
		}
		break;
	}
	return modified;
}

swap(equation, np, level, i1, i2)
token_type	*equation;
int		*np;
int		level;
int		i1, i2;
{
	int	e1, e2;
	int	n1, n2;

	for (e1 = i1 + 1;; e1 += 2) {
		if (e1 >= *np || equation[e1].level <= level)
			break;
	}
	for (e2 = i2 + 1;; e2 += 2) {
		if (e2 >= *np || equation[e2].level <= level)
			break;
	}
	n1 = e1 - i1;
	n2 = e2 - i2;
	blt(&scratch[0], &equation[i1], (e2 - i1) * sizeof(*equation));
	blt(&equation[i1+n2], &equation[e1], (i2 - e1) * sizeof(*equation));
	blt(&equation[i1], &scratch[i2-i1], n2 * sizeof(*equation));
	blt(&equation[e2-n1], &scratch[0], n1 * sizeof(*equation));
}
